# 
# This is a Toy implementation of the General Number Field Sieve For Factoring
# 
# Author: Dan Shumow
# 
# Math 581F Fall 2007
# Professor William Stein
# Final Project 
#
# Math 581E Fall 2008
# Professor Clement Pernet
# Final Project (added block lanczos code.)
#
# Math 583E Spring 2009
# Professor William Stein
# Mini-Project 1 (added lattice sieving code.)
# (also added some more flags, and approximate sieving.)
#
#
# "You haven't lived till you've sieved."
#

#  The following are the sources I used to learn about the General Number Field Sieve

#  Matthew E Briggs Master's Thesis from Virginia Polytech (This is probably the most clear and detailed)
#  The Lenstra et al book "The development of the number field sieve" (esp. for lattice sieve.) 
#  Pomerance's "Tale of Two Sieves"
#  Henri Cohn's "A Course in Computational Algebraic Number Theory"
#  The Math 581F Textbook / Lecture Notes (William Stein)
#  Peter Montgomery's Original Paper on Block Lanczos
#
#  Personal conversations with Peter Montgomery



#
# The General Number Field Sieve works like this
# 
# 1) Decide to use GNFS to factor N, determine algorithm parameters based on N
#    such as, the size of the rational factor base, size of the algebraic factor base
#    number of quadratic characters, degree d, and polynomial f to use
#
# 2) calculate the rational prime factor base (primes up to bound)
# 3) calculate the algebraic factor base (first degree prime ideals up to bound)
# 4) calculate the quadratic characters (first few 1st degree prime ideals beyond the algebraic factor base.)
# 5) use sieving to determine integer pairs (a,b) that have values
#    have a+b*m (m a root of f mod N) and N(a+b*theta) (theta complex root of f)
#    smooth with respect to both the algebraic and rational prime factor bases.
#    while sieving, also determine the degree mod 2 over each prime (rational and algebraic)
#    and save this information.
#    Also, record the power (mod 2) of -1 determining the legendre symbol (a+b*s / q) of the
#    first degree prime ideal (s,q) in the quadratic character base.
# 6) Once there are enough such primes more than the sum of the lengths of the bases
#    use linear algebra to determine the kernel of the transformation.
# 7) Use the result of step 6 to calculate a product of (a+b*theta)'s =  Beta^2 in the number field
# 8) Use the CRT to determine Beta from Beta^2 in the number field
# 9) replace theta in Beta^2 with m to obtain v^2 = phi(m)
# 10) calculate u^2 = product of (a+b*m)'s mod N, calculate u from (a+b*m)
# 11) u^2 - v^2 = 0 mod N, so use difference of squares trick to try to factor N.
# 12) repeat until N is factored.
#


#
# Major Weaknesses of this project:
# The major weaknesses of this project are that I relied a lot on sage
# to do math.  Although it served its purpose in teaching me about the GNFS.
#
# WARNING!!!
# On My Laptop This Sage code can not handle 14 digits.
#
# Other than that:
# 1) To calculate square roots, I should use the CRT, but instead
#    I let sage functions do the work for me.
#    I did this because, the CRT is a technical detail at the end of the
#    algorithm, and also, I understand that math already.
#    I wanted to focus on learning the stuff I didn't understand.
# 2) I rely on sage for all linear algebra, this wouldn't work at limits where
#    10s or 100s of thousands of elements may be in the factor base.
# 3) This program uses is_prime_power() to try and determine if an N is a prime power
#    (because the GNFS does not work well on prime powers.)
# 4) If a number is a prime power, I use the ECM to factor it, in actuality
#    an algorithm would be selected heuristically.
# 5) The linear algebra code is the primary weak spot in the sage code at this time.
#

#
# Global Flags: These can be set to change the behavior of the sieving algorithm
# Used for sort of anecdotal observation and experimentation of the algorithm
#

# if this value is set to false, the algorithm will not use quadratic characters
use_quadratic_characters = True;

# if this value is set to true, the algorithm uses logarithms to sieve approximately.
use_approximate_sieving = True;

# if this value is set to true, the algorithm uses lattice sieving to sieve
use_lattice_sieving = True;

# if this value is set to true, the algorithm uses sparse linear algebra
use_sparse_linalg = False;

#
# This cell implements simple heuristics for the
# parameters used in the GNFS based on given N
#
# These parameters are really rough because analysing how to run the 
# sieve at the small sizes (that the sage linear algebra can handle)
# is really hard.
#

rational_factor_base_bound_ratio = 2.0;    # this factor seemed to work well
algebraic_factor_base_bound_ratio = 4.0;  # so did this
quadratic_character_base_ratio = 0.10;   # as a percentage of the algebraic factor base

degree = 3;     # hard coded in this toy implementation

a_bound_ratio = 5.0;

log_bound = 10.0;

def determine_GNFS_bounds(N):
    B1 = max(10, ceil(rational_factor_base_bound_ratio*log(N)));
    B2 = max(20, ceil(algebraic_factor_base_bound_ratio*log(N)));

    if (use_quadratic_characters):
        Quad_character_count = max(10, ceil((1.0+quadratic_character_base_ratio)*B2))
    else:
        Quad_character_count = 0

    B3 = B2 + Quad_character_count;

#    a_bound = min(max(50, ceil(a_bound_ratio * log(N))), (2^16 - 1) );

    a_bound = max(50, ceil(a_bound_ratio * log(N)));

    C_bound = ZZ(round(sqrt(a_bound)));
    D_bound = C_bound;

    print 'Automatic Parameter Selection for GNFS:'

    print 'Small prime bound (B1):', B1
    print 'Medium prime bound (B2):', B2
    print 'Large prime bound (all subsequent primes under):', B3

    print 'search bounds for unit term: [', -a_bound, ',', a_bound, ']'

    return (B1, B2, B3, degree, a_bound, C_bound, D_bound);



#
# Note about the representation of prime base elements
# rational primes are represented as (m mod p, p) p prime in QQ
# algebraic primes are represented as (r, p) p prime in QQ, r root of f mod p.
# storing in this form helps sieving and allows common code to be used
# for both
#

# The function that
# generates the rational factor base
# (rational primes for sieving)

def generate_rational_factor_base(m, Bound):

    rational_factor_base = [];

    for p in xrange(1,Bound):
        if(is_prime(p)):
            rat_fact = ( m%p, p); 
            if (use_approximate_sieving):
                rat_fact = ( rat_fact[0], rat_fact[1], ZZ(rat_fact[1]).nbits() )
            rational_factor_base.append( rat_fact );

    return rational_factor_base;



# This cell generate the algebraic factor base
# (first degree prime ideals of order Z[theta] in number field)

# helper functions:

# uses sage to determine roots of f mod p
def get_roots_mod_p(f, p):
    return f.change_ring(GF(p)).roots(None, False);

# generates the algebraic factor base
def generate_algebraic_factor_base(f, Bound):

    algebraic_factor_base = [];

    for p in xrange(Bound):
        if(is_prime(p)):

            roots_fmodp = get_roots_mod_p(f, p);
            log_p = ZZ(p).nbits();

            for j in xrange(len(roots_fmodp)):
                alg_fact = ( roots_fmodp[j].lift(), p );
		if (use_approximate_sieving):
                    alg_fact = (alg_fact[0], alg_fact[1], log_p);
                algebraic_factor_base.append( alg_fact )
                
    return algebraic_factor_base;

# generates quadratic characters
def generate_quadratic_character_base(f, Lower_Bound, Upper_Bound):

    quadratic_character_base = [];

    for p in xrange(Lower_Bound, Upper_Bound):
        if(is_prime(p)):
            roots_fmodp = get_roots_mod_p(f, p);

            for j in xrange(len(roots_fmodp)):
                quadratic_character_base.append((roots_fmodp[j].lift(),p))
                
    return quadratic_character_base;



#
# this cell defines helper functions for polynomials
# 
#

#
# determines m-arry expansion of N
# and the coefficients of powers of m
def get_polynomial_coefficients(N, m):

    coefficient_list = [];

    temp = N;

    while(0 != temp):
        r = temp % m;
        temp = (temp - r) / m;
        coefficient_list.append(r);    

    return coefficient_list;
    


#
# function that tries to generate a
# monic, irreducible polynomial f with root m mod N
# with degree d.
# 
def generate_polynomial(polynomial_ring, N, d):
    # determine and lower and upper bounds for m

    q = 0;

    while (1 != q):

        u = floor(pow(float(N), 1.0/d));
        l = ceil(pow(float(N), (2.0*d - 3.0)/(2.0*d*(d-1))));
    
        m = floor(round(uniform(l, u)));

        (q,r) = ZZ(N).quo_rem(m^d);
    

    f = polynomial_ring(get_polynomial_coefficients(N, m));
   
    return (f, m)



#
# checks if a polynomial is monic and irreducible
#
def is_good_polynomial(f):
    return (f.is_monic() and f.is_irreducible())



#
# retries a bunch of times to generate a suitable random polynomial
#
def get_polynomial(polynomial_ring, N, d, retry_count):

    found_polynomial = False;

    for j in xrange(retry_count):
        (f, m) = generate_polynomial(polynomial_ring, N, d);
        if is_good_polynomial(f):
            found_polynomial = True;
            break;

    if (found_polynomial):
        return (f, m)
    else:
        return (0,0)



#
#    This cell contains the sieving code
#

quo_rem = Integer.quo_rem


#
#    Common code that will take a factor base and sieve it
#
def sieve_interval(interval_to_sieve, a_lowerbound, b, factor_base, approx_bound=None, approx_interval=None):


    factor_base_count = len(factor_base);
    
    exponent_lists = [ [0 for j in xrange(factor_base_count+1)] for k in xrange(len(interval_to_sieve))]

    first_phase = 0;
    if (not use_approximate_sieving):
        first_phase = 1;

    for phase in xrange(first_phase,2):

        #
        #    do the factor base sieving
        #

        #
        #    determine the sign exponents of the elements in each interval
        # 
        if (1 == phase) :
            for j in xrange(len(interval_to_sieve)):
                if ( use_approximate_sieving and (approx_bound < approx_interval[j]) ):
                    continue
                if (interval_to_sieve[j] < 0):
                    exponent_lists[j][factor_base_count] = 1;
                    interval_to_sieve[j] = abs(interval_to_sieve[j]);

        for j in xrange(factor_base_count):
            p = factor_base[j][1];
            bt = -b*factor_base[j][0] % p;

	    if (0 == phase):
                log_p = factor_base[j][2];
            
            #
            #    Much of the power of the sieving algorithms
            #    come from the regular access pattern and
            #    advancing through the array p elements at a time
            #
    
            # in this case we have our factor bases stored as a pair (t,p)
            # such that the value in the sieve in that position
            # (for the rational base this is a+b*m and for the algebraic base it is N(a+b*theta))
            # is only divisible by p if a = -b*t + k*p, as such we
            # use the lower bound of the a range to determine the first
            # position in the array to start sieving at advance p positions from there
    
            for k in xrange((bt  - a_lowerbound) % p, len(interval_to_sieve), p):
    
                if (0 == phase):
                    approx_interval[k] = approx_interval[k] - log_p;
                else:

                    if use_approximate_sieving:
                        if (approx_bound < approx_interval[k]):
                            continue;

                    #
                    # This loop goes through and calculates the largest integer e
                    # such that p^e divides the array entry
                    #
                
                    e = 0;
                    r = 0;
    
                    while (0 == r):
                        if (interval_to_sieve[k] != 0):
                            (q,r) = quo_rem(interval_to_sieve[k] , p);
                            if(0 == r):
                                r = 0;
                                e = e + 1;
                                interval_to_sieve[k] = q;
                            else:
                                r = 1;
                        else:
                            r = 1;

                    exponent_lists[k][j] = e;

    return (interval_to_sieve, exponent_lists);



#
# determines the quadratic character exponent vector
#
def determine_quad_char_expons(a,b, quad_char_base):

    qc_expons = [0 for j in xrange(len(quad_char_base))];

    for j in xrange(len(quad_char_base)):

        q = quad_char_base[j][1];
        s = quad_char_base[j][0];        

        qc_expons[j] = Integer((1 - legendre_symbol(a+s*b, q))/2);

    return qc_expons;

#
#   function: search_for_smooth_pairs_interval
#   input: 6 parameters:
#   lower bound of a's
#   b
#   list of a+bm, sieved with respect to rational factor base
#   The exponent vector with respect to the rational factor base
#   list of N(a+b*theta) sieved with respect to the algebraic factor base
#   The exponent vector with respect to the algebraic factor base

#   output: a list of 4-tuples of:
#   (a,b) pairs of integers that are smooth with respect to
#         both the algebraic and rational factor bases.
#   the exponent list with respect to the rational factor base
#   the exponent list with respect to the algebraic factor base
#   the exponent list with respect to the quadratic character base
#

def search_for_smooth_pairs_interval(a_lowerbound, b, rat_sieve, rat_exp_list, alg_sieve, alg_exp_list, quad_char_base):

    smooth_list = [];

    for j in xrange(len(rat_sieve)):
        if ( (1 == rat_sieve[j]) and (1 == alg_sieve[j]) ):
            
            qc_expons = determine_quad_char_expons(a_lowerbound+j, b, quad_char_base);

            smooth_list.append(((a_lowerbound+j,b),rat_exp_list[j],alg_exp_list[j], qc_expons));

    return smooth_list;



#
# Do the actual sieving
# returns a list of (a,b) pairs along with the rational, algebraic, and quadratic exponents
#

def perform_interval_sieving(f, m, a_bound, rat_factor_base, alg_factor_base, quad_char_base):

    # determine the dimension of the space we are working over.    
    l1 = len(rat_factor_base);
    l2 = len(alg_factor_base);
    l3 = len(quad_char_base);

    l = l1 + l2 + l3 + 2; # add two for rational and algebraic sign exponents;

    d = f.degree()

    smooth_element_list = [];

    rat_log_bound = log_bound*rat_factor_base[l1-1][1];
    alg_log_bound = log_bound*alg_factor_base[l2-1][1];

    # loop until we have more smooth elements
    # than we have dimension

    b = 1;

    while(len(smooth_element_list) <= l):

        rat_interval = [];
        alg_interval = [];

        approx_rat_interval = None;
        approx_alg_interval = None;

        if (use_approximate_sieving):
            approx_rat_interval = [];
            approx_alg_interval = [];
    
        for j in xrange(2*a_bound + 1):

            a = -a_bound + j;

            rat_norm = a + b*m; # sieve the value a + b*m
            alg_norm = Integer(((-b)^d)*f(-a/b)); # sieve the value N(a+b*theta) = (-b)^d * f(-a/b)

            rat_interval.append(rat_norm);
            alg_interval.append(alg_norm);

            if (use_approximate_sieving):
                if (0 != rat_norm): 
                    log_rat_norm = rat_norm.nbits();
                else:
                    log_rat_nrom = infinity;

                if (0 != alg_norm):
                    log_alg_norm = alg_norm.nbits();
                else:
                    log_alg_norm = infinity;

                approx_rat_interval.append(log_rat_norm);
                approx_alg_interval.append(log_alg_norm);
        
        (rat_interval, rat_exp_list) = sieve_interval(rat_interval, -a_bound, b, rat_factor_base, \
                                           approx_bound=rat_log_bound, approx_interval=approx_rat_interval);
        
        (alg_interval, alg_exp_list) = sieve_interval(alg_interval, -a_bound, b, alg_factor_base, \
                                           approx_bound=alg_log_bound, approx_interval=approx_alg_interval);

        smooth_pairs_found = search_for_smooth_pairs_interval(-a_bound, b, rat_interval, rat_exp_list, alg_interval, alg_exp_list, quad_char_base);
 
        smooth_element_list.extend(smooth_pairs_found);

        print 'found', len(smooth_element_list), 'simultaneously smooth pairs.'

        b = b+1;

    return smooth_element_list;



#
#    This cell contains a function that computes the product of the integer pairs
#    as an integer polynomial, and reduces it modulo the ideal generated by f
#

#
# computes product of linear polynomials mod ideal generated by f
# (equivalent to doing math in number field)
#
def compute_numberfield_product_of_pairs(polynomial_ring, f, integer_pairs, vector):

    I = polynomial_ring.ideal(f);

    product_polynomial = polynomial_ring([1]);

    for j in xrange(len(vector)):

        if (1 == vector[j]):

            linear_poly = polynomial_ring([integer_pairs[j][0], integer_pairs[j][1]]);
            product_polynomial = product_polynomial * linear_poly;
            product_polynomial.mod(I);

    return product_polynomial;



#
# computes the product of (a + b*m)'s mod N
# of the (a + b*m)'s corresponding to 1s in vector
#
def compute_integer_product_of_pairs(polynomial_ring, f, m, N, integer_pairs, vector):

    prod  = 1;

    for j in xrange(len(vector)):
        if (1 == vector[j]):
            prod  = prod*(integer_pairs[j][0] + m*integer_pairs[j][1]) % N;

    return prod;


    
#
# computes the difference of squares u^2 - v^2
# and returns (u+v) and (u-v)
#
def compute_difference_of_squares(polynomial_ring, f, m, N, integer_pairs, vector):

    found_squares = False;

    u_plus_v = 0; 
    u_minus_v = 0;

    ints_mod_N = ZZ.quo(ZZ.ideal(N));

    vsquared = ints_mod_N(compute_integer_product_of_pairs(polynomial_ring, f, m, N, integer_pairs, vector));

    if ( is_square(vsquared)):

        beta_squared = compute_numberfield_product_of_pairs(polynomial_ring, f, integer_pairs, vector);
        beta = find_square(f, beta_squared, polynomial_ring);
    
        if (False != beta):
        
            u = (beta(m)) % N;
            v = vsquared.sqrt().lift();
            u_plus_v = (u + v) % N; 
            u_minus_v = (u - v) % N;

            found_squares = True;

        else:
            print 'failed to find a square root in number field.'
    else:
        print 'integer was not square'

    return (found_squares, u_plus_v, u_minus_v);



#
#    Cheater Helper Functions
#
#    These are functions that I was too lame to implement myself
#    And Hence relied on sage's built in functionality.
#

ecm_instance = ECM();

def factor_prime_power(N):
    # The GNFS doesn't work one prime powers
    # so we use ecm to factor,
    # in reality, 
    # we would apply a heuristic on the best way to determine the prime and exponent    
    
    ecm_factors = ecm_instance.factor(N);

    prime_power_factorization = [(ecm_factors[0], len(ecm_factors))]

    return Factorization(prime_power_factorization);

#
#    This simple function does all the linear algebra 
#    It works by calling my Block Lanczos function
#    (my project for Math 581E Fall 08)
#
def find_kernel_vectors(M):
    return FindDependency_by_BlockLanczos(M.transpose(),16).transpose().rows();

#
# In reality this would be done with the CRT.
#
def find_square(f, beta_squared, polynomial_ring):
    # computes the square root of beta_squared
    # (a polynomial over ZZ of degree less than d, that represents the element in the order ZZ[theta])
    # and returns the polynomial representing the square root.
    
    K.<a> = NumberField(f);
    beta2 = beta_squared(a);
    if(not beta2.is_square()):
        print 'The supposedly found square is not square!'
        return False;
    return polynomial_ring(beta2.sqrt().list());


def GNFS_Sieving_Stage(f, m, a_bound, C_bound, D_bound, B1, B2, B3):

    # get the rational factor base
    rational_factor_base = generate_rational_factor_base(m, B2);

    # get the algebraic factor base
    algebraic_factor_base = generate_algebraic_factor_base(f, B2);

    # get the quadratic character base
    quadratic_character_base = generate_quadratic_character_base(f, B2, B3);

    #
    # get a list of smooth elements
    #

    if (use_lattice_sieving):
        smooth_element_list = perform_lattice_sieve(f, m, B1, B2, C_bound, D_bound, rational_factor_base, algebraic_factor_base, quadratic_character_base);
    else:
        smooth_element_list = perform_interval_sieving(f, m, a_bound, rational_factor_base, algebraic_factor_base, quadratic_character_base);

    return smooth_element_list;



#
# This is the factoring algorithm
# It factors N, and returns a sage factorization object of N
#

def GNFS_Factoring_Algorithm(N, f=None, m=None, B1=None,B2=None,B3=None,a_bound=None,C_bound=None,D_bound=None,override=False):

    # determine if this is a good number to use the GNFS on
    # if it is not (i.e. it is a prime power)
    # call the specialized prime power factorization function

    if(is_prime_power(N)):

        if (is_prime(N)):
            return factor(N);

        print N, 'is a prime power, reducing to using ECM to factor'

        return factor_prime_power(N);

    # determine the algorithm bounds
    if (not override):
        (B1, B2, B3, d, a_bound, C_bound, D_bound) = determine_GNFS_bounds(N);

        R.<x> = ZZ[];

        (f, m) = get_polynomial(R, N, d, 20);

        if((0 == f) and (0 == m)):
    
            # if the polynomial selection code did not find a degree d
            # polynomial in 20 tries, just use the built-in factoring
            # algorithm, because N is too small for the GNFS to work
            # at that point

            return factor(N);

    print 'polynomial =', f

    smooth_element_list = GNFS_Sieving_Stage(f, m, a_bound, C_bound, D_bound, B1, B2, B3);

    if len(smooth_element_list) <= 1:
        print 'totally failed to find a sufficient number of smooth elements.  retry with different parameters.'
        return;

    #
    #    parse out the list of smooth elements into a list of integer pairs,
    #

    integer_pairs = [smooth_element_list[k][0] for k in xrange(len(smooth_element_list))];
    exponent_vectors = [];

    #
    # use the smooth element list vectors to make a matrix
    #

    for k in xrange(len(smooth_element_list)):
        exponent_vectors.extend(smooth_element_list[k][1]);
        exponent_vectors.extend(smooth_element_list[k][2]);
        exponent_vectors.extend(smooth_element_list[k][3]);

    num_expons = len(smooth_element_list[0][1]) + len(smooth_element_list[0][2]) + len(smooth_element_list[0][3]);
        

    M = matrix(GF(2), len(smooth_element_list), exponent_vectors, sparse=use_sparse_linalg)
    
    solutions = find_kernel_vectors(M);

    print "Linear Algebra step found:", len(solutions), "vectors in kernel space."

    nontrivial_first_factor = False;
    nontrivial_second_factor = False;

    k = 0

    #
    #    Loop over the vectors that sum to zero to try and find one
    #    That leads to a non trivial factorization
    #
    while ((k < len(solutions)) and (not nontrivial_first_factor) and (not nontrivial_second_factor)):
        
        (found_squares, u_plus_v, u_minus_v) = compute_difference_of_squares(R, f, m, N, integer_pairs, solutions[k]);

        if(found_squares):
                
            first_factor_of_N = gcd(u_plus_v, N);

            second_factor_of_N = gcd(u_minus_v, Integer(N/first_factor_of_N));

            if((1 != first_factor_of_N) and (N != first_factor_of_N)):
                nontrivial_first_factor = True;
            if((1 != second_factor_of_N) and (N != second_factor_of_N)):
                nontrivial_second_factor = True;
            else:
                print 'failed to find a nontrivial factor.'
    
        else:
            print 'failed to find squares.'

        k = k + 1;


    factors_of_N = [];

    N_remaining = N;

    if(nontrivial_first_factor):
        factors_of_N.append(first_factor_of_N);
        N_remaining = Integer(N_remaining/first_factor_of_N)

    if(nontrivial_second_factor):
        remaining_second_factor = gcd(second_factor_of_N, N_remaining);
        factors_of_N.append(remaining_second_factor);
        N_remaining = Integer(N_remaining/second_factor_of_N);

    if(1 != N_remaining):
        factors_of_N.append(N_remaining);

    factorization_of_N = factor(1);

    #
    #    Recursively factor non trivial factors
    #    Or N, if we completely failed to find a factor
    #

    for j in xrange(len(factors_of_N)):
        factorization_of_N = factorization_of_N * GNFS_Factoring_Algorithm(factors_of_N[j]);

    return factorization_of_N;



#
#    Factoring test code
#    This function will randomly generate an integer and factor it with the GNFS
#
#    It fails sometimes, due to stuff like low memory conditions, and bugs
#

def test_GNFS():

    N = Integer(ceil(uniform(1, 1000000)));

    print 'Factoring N =', N

    GNFS_Factorization = GNFS_Factoring_Algorithm(N);    
    
    if (GNFS_Factorization.value() != N):
        print 'Test Case N =', N, ' FAILED!!!'
        print 'says factorization is:', GNFS_Factorization
        return False

    print GNFS_Factorization

    print 'Test Case PASSED!'

    return True

#
# Linear Algebra Code
# Algorith: Block Lanczos
#

def BlockLanczos(B, Y):

    #
    # Initialize Variables
    #

    F = B.base_ring()

    n1 = B.nrows();
    n2 = B.ncols();

    # assert correct dimensions
    if (n2 != Y.nrows()):
        print 'Y does not have correct number of rows.'
        return;

    N = Y.ncols();

    # initialize previous W matrices
    W_inv_2 = matrix(F, N, N, sparse=use_sparse_linalg);
    W_inv_1 = matrix(F, N, N, sparse=use_sparse_linalg);

    # initialize previous V matrices
    V_2 = matrix(F, n2, N, sparse=use_sparse_linalg);
    V_1 = matrix(F, n2, N, sparse=use_sparse_linalg);

    Cond_1 = matrix(F, N, sparse=use_sparse_linalg);
    K_1 = matrix(F, N, sparse=use_sparse_linalg);

    #
    # initialize temporary product matrices
    #
    B_V_1 = matrix(F, n1, N, sparse=use_sparse_linalg);
    V_trans_B_trans_1 = matrix(F, N, n1, sparse=use_sparse_linalg);
    S_1 = I_N = identity_matrix(F, N, sparse=use_sparse_linalg);
    X = matrix(F, n2, N, sparse=use_sparse_linalg);
    
    #
    # calculate temporary product matrices
    #

    # initialize the current Vi matrix
    V_0 = V_i = B.transpose()*(B*Y);
    
    # initialize the product of B and V_i
    B_V_i = B*V_i;

    # initialize the transpose of the product of V and B
    V_trans_B_trans_i = B_V_i.transpose();

    # initialize the conditional matrix
    Cond_i = V_trans_B_trans_i * B_V_i;

    i = 0;

    while (not Cond_i.is_zero()):

        if (i > n2):
            print ' bailing because i is to big ... '
            break;

        # Generate the W and S matrices    
        (W_inv_i, S_i) = Generate_W_and_S(Cond_i, S_1, N, F);

        X = X + V_i * (W_inv_i * (V_i.transpose() * V_0));
        
        B_trans_B_V_i = B.transpose() * B_V_i;
        K_i = (V_trans_B_trans_i * (B * B_trans_B_V_i))* S_i + Cond_i;
        
        D_i = I_N - W_inv_i*K_i;
        
        E_i = -W_inv_1*Cond_i*S_i;

        F_i = -W_inv_2*(I_N - Cond_1*W_inv_1)*K_1*S_i;

        V_next_i = B_trans_B_V_i * S_i + V_i*D_i + V_1*E_i + V_2*F_i;

        B_V_i = B*V_next_i;
        
        V_trans_B_trans_i = B_V_i.transpose();

        Cond_next_i = V_trans_B_trans_i*B_V_i;
        
        i = i + 1;
        
        #
        # Swap the previous state to the next previous state
        #
        W_inv_2 = W_inv_1;
        V_2 = V_1;

        #
        # Swap the current state to the previous state
        #
        S_1 = S_i
        W_inv_1 = W_inv_i;
        V_1 = V_i;
        Cond_1 = Cond_i;
        K_1 = K_i;

        #
        # Swap the next state to the current state
        #
        V_i = V_next_i;
        Cond_i = Cond_next_i;
        
    return (X, V_i);



def Generate_W_and_S(T, S_1, N, F):

    M = T.augment(identity_matrix(F, N,sparse=use_sparse_linalg));
    
    S = [];

    #
    #  First we have to reorder the columns of T
    #

    c = [];
    
    for j in xrange(N):
        if (0 == S_1[j,j]):
            c.append(j);
    
    for j in xrange(N):
        if (1 == S_1[j,j]):
            c.append(j);

    for j in xrange(N):
        
        cj = c[j];
        
        for k in xrange(j,N):
            
            ck = c[k]
            
            if (0 != M[cj,cj]):
                break;
            
            if (0 != M[ck, cj]):
                M.swap_rows(ck, cj);
        
        if (0 != M[cj,cj]):
            S.append(cj);
            M.set_row_to_multiple_of_row(cj,cj, 1/M[cj,cj]);
            for k in xrange(N):
                ck = c[k];
                if(j != k):
                    M.add_multiple_of_row(ck, cj, -M[ck,cj]);
        else:
            # then we did not find a pivot in column j;
            # Thus column j is a linear combination of previous columns
            #  skip it.
            for k in xrange(j,N):
                
                ck = c[k]

                if (0 != M[cj, cj+N]):
                    break;
                
                if (0 != M[ck, cj+N]):
                    M.swap_rows(cj, ck);

            if (not (0 != M[cj, cj+N])):
                print 'ASSERT(0 != M[cj,cj+N]) FAILED!'
                #print S
                #print c
                #print j, cj
                #print M
                #print T
                #print S_1
            
            for k in xrange(j, N):
                ck = c[k];
                if (j != k):
                    M.add_multiple_of_row(ck, cj, -M[ck,cj+N]);
            
            M.set_row_to_multiple_of_row(cj,cj, 0);
    

    W_inv_i = M.submatrix(0,0,M.nrows(),N);
    S_i = matrix(F, N, N, sparse=use_sparse_linalg);
    for j in xrange(len(S)):
        S_i[S[j],S[j]] = 1;

    return (W_inv_i, S_i);



def FindDependency_by_BlockLanczos(B, N):

    n2 = B.ncols();

    #
    # Generate Y and make sure the columns
    # are not A self-orthogonal.
    #
    
    Y = random_matrix(B.base_ring(), n2, N,sparse=use_sparse_linalg);

    for k in xrange(N):
        while (0 == Y.column(k)*B.transpose()*B*Y.column(k).transpose()):
            Y.set_column(k, random_matrix(B.base_ring(), n2, 1, sparse=use_sparse_linalg).list());
    
    [X, Vm] = BlockLanczos(B, Y);
    
    Z = (X-Y);
    Z = Z.augment(Vm);

    B_Z = B*Z;

    U = B_Z.transpose().kernel().basis_matrix().transpose();

    Z_U = Z*U;

    return Z_U;


def get_short_basis_vectors(t, p):
    M = matrix(ZZ, [[-t, 1], [p,0]]);
    M = M.LLL();

    v1 = (M[0,0], M[0,1]);
    v2 = (M[1,0], M[1,1]);

    return (v1, v2);


def lattice_sieve_by_row(lattice_to_sieve, approx_lattice, exponent_list, C_bound, D_bound, approx_bound, p_index, t, p, log_p, v1, v2, phase):
    
    u1 = (v1[0] + t*v1[1]) % p;
    u2 = (v2[0] + t*v2[1]) % p;

    u2_is_unit = True;

    if ( 0 != u1 ) and ( 0 != u2 ):

        u2_inv = xgcd(u2,p)[1];
        delta = (-u1*u2_inv) % p

    else:
        delta = 0;

        if (0 == u2):
            u2_is_unit = False;

    C = -C_bound;

    start_pos = (C*delta) % p;
    if 0 == start_pos:
        start_pos = p;

    while (C < C_bound):

        D = start_pos;
        row_advance = p;

        if not u2_is_unit:
            if 0 == (C % p):
                row_advance = 1;
                D = 1;
            else:
                D = D_bound;

        while (D < D_bound):
            j = C + C_bound;
            k = D-1;

            if use_approximate_sieving and (0 == phase):
                approx_lattice[j][k] -= log_p;
            elif (not use_approximate_sieving) or (approx_lattice[j][k] < approx_bound):
                e = 0;
                x = lattice_to_sieve[j][k]
                (q,r) = quo_rem(x,p);
 
                if (0 != x):
                    while (0 == r):
                        e =+ 1;
                        x = q;
                        (q,r) = quo_rem(x,p);

                lattice_to_sieve[j][k] = x;
                exponent_list[j][k][p_index] = e;

            D = D + row_advance;

        start_pos = (start_pos + delta) % p;
        C = C+1;

    return;


#
# This function generates the short vectors for the lattice of the ideal (pq)
#
def get_pq_vectors(p, s, q, v1, v2):
    
    # the 'residue characteristic' of (pq) [note that this is not truly a res. char.]
    pq = p*q;
    
    # The short vectors that generate (pq)
    L = Matrix(ZZ, [[-p*s, p],[pq,0]]).LLL();

    # compute the short vectors in the (c,d) plane

    M = Matrix(ZZ, [[v1[0],v2[0],L[0,0]],[v1[1],v2[1],L[0,1]]]).echelon_form();
    u1 = (M[0,2],M[1,2]);

    M = Matrix(ZZ, [[v1[0],v2[0],L[1,0]],[v1[1],v2[1],L[1,1]]]).echelon_form();
    u2 = (M[0,2],M[1,2]);

    return (u1, u2);


def lattice_sieve_by_vector(lattice_to_sieve, approx_lattice, exponent_list, C_bound, D_bound, approx_bound, p_index, t, p, log_p, v1, v2, s, q, phase):

    (u1, u2) = get_pq_vectors(p, s, q, v1, v2);

    #
    # These bounds were suggested in  Pollard's paper.
    #

    for e in xrange(-2,2):

        for f in xrange(1,2):

            c = e*u1[0] + f*u2[0];
            j = c + C_bound;

            if (j < 0) or (len(lattice_to_sieve) <= j):
                continue;
           
            d = e*u1[1] + f*u2[1];
            k = d - 1;

            if (k < 0) or (len(lattice_to_sieve[j]) <= k):
                continue;

            if use_approximate_sieving and (0 == phase):
                approx_lattice[j][k] -= log_p;
            elif (not use_approximate_sieving) or ((1==phase) and (approx_lattice[j][k] <= approx_bound)):
                print 'Doing Exact Sieving'
                e = 0;
                x = lattice_to_sieve[j][k]
                (q,r) = quo_rem(x,p);
 
                if (0 != x):
                    while (0 == r):
                        e =+ 1;
                        x = q;
                        (q,r) = quo_rem(x,p);
                if (0 == e) and not (0 == x):
                    print 'expected divisibility was not divisible'
                else:
                    print 'got divisibility where we expected it'
                lattice_to_sieve[j][k] = x;
                exponent_list[j][k][p_index] = e;

    return;                


def sieve_lattice(lattice_to_sieve, C_bound, D_bound, B1, factor_base, approx_bound, approx_lattice, v1, v2, s, q):

    #
    # At the sizes we are using for this toy implementation
    # sieving by vectors just does not work
    # so we set this bound to infinity
    #
    lattice_sieve_by_vector_bound = infinity;
    
    factor_base_count = len(factor_base);

    exponent_lists = [ [ [0 for l in xrange(factor_base_count+1)] for j in xrange(1,D_bound+1)] for k in xrange(-C_bound, C_bound+1)];

    first_phase = 0;
    if (not use_approximate_sieving):
        first_phase = 1;

    for phase in xrange(first_phase,2):
        #
        # do the factor base sieving
        #

        if (1 == phase) :
            for j in xrange(len(lattice_to_sieve)):
                for k in xrange(len(lattice_to_sieve[j])):
                    if ( use_approximate_sieving and (approx_bound < approx_lattice[j][k]) ):
                        continue;
                    if (lattice_to_sieve[j][k] < 0):
                        exponent_lists[j][k][factor_base_count] = 1;
                        lattice_to_sieve[j][k] = abs(lattice_to_sieve[j][k]);

        for k in xrange(factor_base_count):
		    
            t = factor_base[k][0];
            p = factor_base[k][1];
                    
            log_p = None;
            if (use_approximate_sieving):
                log_p = factor_base[k][2];

            if (p < lattice_sieve_by_vector_bound):
                lattice_sieve_by_row(lattice_to_sieve, approx_lattice, exponent_lists, C_bound, D_bound, approx_bound, k, t, p, log_p, v1, v2, phase);
            else:
                lattice_sieve_by_vector(lattice_to_sieve, approx_lattice, exponent_lists, C_bound, D_bound, approx_bound, k, t, p, log_p, v1, v2, s, q, phase);

    return (lattice_to_sieve, exponent_lists);


def search_for_smooth_pairs_lattice(rat_lattice, alg_lattice, alg_exp_lists, rat_exp_lists, quad_char_base, C_bound, D_bound, v1, v2, smooth_pairs):

    for C in xrange(-C_bound, C_bound):
        for D in xrange(1,D_bound):
            j = C_bound + C;
            k = D-1;
            if (rat_lattice[j][k] == 1) and (alg_lattice[j][k] == 1):
                a = C*v1[0] + D*v2[0];
                b = C*v1[1] + D*v2[1];
                
                qc_expons = determine_quad_char_expons(a, b, quad_char_base);
                
                smooth_pairs.append( ( (a,b), rat_exp_lists[j][k], alg_exp_lists[j][k], qc_expons ) );

    return;


def perform_lattice_sieve(f, m, B1, B2, C_bound, D_bound, rat_factor_base, alg_factor_base, quad_char_base):

    # determine the dimension of the space we are working over.    
    l1 = len(rat_factor_base);
    l2 = len(alg_factor_base);
    l3 = len(quad_char_base);

    l = l1 + l2 + l3 + 2; # add two for rational and algebraic sign exponents;

    d = f.degree()
    
    smooth_element_list = [];
    
    j = l2-1;
    
    approx_rat_bound = log_bound*(ZZ(rat_factor_base[l1-1][1]).nbits());
    approx_alg_bound = log_bound*(ZZ(alg_factor_base[l2-1][1]).nbits());

    #
    #  Loop until we have enough smooth elements
    #
    while (len(smooth_element_list) <= l):
        
        if (j < 0):
            print 'underflowed algebraic prime list, try different parameters'
            break;

	#
	# we need to find short vectors in the lattice formed by each ideal
        # so we use LLL (Pollard spoke of another method, but I couldn't figure it out)
	#
        s = alg_factor_base[j][0];
        q = alg_factor_base[j][1];

        if (q < B1):
            print 'exhausted all midrange primes and did not find enough smooth elements'
            break;

        (v1, v2) = get_short_basis_vectors(s, q);

        rat_lattice = [];
        alg_lattice = [];

        approx_rat_lattice = None;
        approx_alg_lattice = None;

        if (use_approximate_sieving):
            approx_rat_lattice = [];
            approx_alg_lattice = [];

        for C in xrange(-C_bound, C_bound):
            rat_row = [];
            alg_row = [];

            if (use_approximate_sieving):
                approx_rat_row = [];
                approx_alg_row = [];

            for D in xrange(1, D_bound):

                a = C*v1[0] + D*v2[0];
                b = C*v1[1] + D*v2[1];

                rat_norm = a + b*m; # sieve the value a + b*m
                if (b != 0):
                    alg_norm = Integer(((-b)^d)*f(-a/b)); # sieve the value N(a+b*theta) = (-b)^d * f(-a/b)
                else:
                    alg_norm = (-a)^d
          
                if (use_approximate_sieving):
                    if (0 != rat_norm): 
                        log_rat_norm = rat_norm.nbits();
                    else:
                        log_rat_norm = infinity;

                    if (0 != alg_norm):
                        log_alg_norm = alg_norm.nbits();
                    else:
                        log_alg_norm = infinity;

                    approx_rat_row.append(log_rat_norm);
                    approx_alg_row.append(log_alg_norm);        
                
                rat_row.append(rat_norm);
                alg_row.append(alg_norm);

                if (use_approximate_sieving):
                    approx_rat_row.append(log_rat_norm);
                    approx_alg_row.append(log_alg_norm);
                
            rat_lattice.append(rat_row);
            alg_lattice.append(alg_row);

            if (use_approximate_sieving):
                approx_rat_lattice.append(approx_rat_row);
                approx_alg_lattice.append(approx_alg_row);

        (rat_lattice, rat_exponents) = sieve_lattice(rat_lattice, C_bound, D_bound, B1, rat_factor_base, approx_rat_bound, approx_rat_lattice, v1, v2, s, q);

        (alg_lattice, alg_exponents) = sieve_lattice(alg_lattice, C_bound, D_bound, B1, alg_factor_base, approx_alg_bound, approx_alg_lattice, v1, v2, s, q);
        
        search_for_smooth_pairs_lattice(rat_lattice, alg_lattice, rat_exponents, alg_exponents, quad_char_base, C_bound, D_bound, v1, v2, smooth_element_list);

        print 'Lattice Sieving found', len(smooth_element_list), 'smooth pairs.'

        j = j-1;

    return smooth_element_list;